#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mymt.h"
#include "cpu_x86.h"

typedef unsigned int uint32;

void (*myMT_Next)(myMT_t *mt);

/* Period parameters */  
#define N 624
#define M 397
#define MATRIX_A 0x9908B0DF   /* constant vector a */
#define UMASK 0x80000000 /* most significant w-r bits */
#define LMASK 0x7FFFFFFF /* least significant r bits */
#define MIXBITS(u,v) ( ((u) & UMASK) | ((v) & LMASK) )
#define TWIST(u,v) ((MIXBITS(u,v) >> 1) ^ ((v)&1 ? MATRIX_A : 0))


/////////////////////////////////////////////////////////////////////////

/* defined in mymt_x86.nas */
void myMT_Next_SSE2(myMT_t *mt);
void myMT_Next_MMX(myMT_t *mt);
//void myMT_Next_other(myMT_t *mt);

void myMT_Next_C(myMT_t *mt)
{
	int i;
	uint32 *p, *q;

	p = mt->state;
	for (i = 0; i < N - M; i++, p++) {
		*p = p[M] ^ TWIST(p[0], p[1]);
	}

	for (i = 0; i < M - 1; i++, p++) {
		*p = p[M-N] ^ TWIST(p[0], p[1]);
	}

	*p = p[M-N] ^ TWIST(p[0], mt->state[0]);
	mt->index = 0;

	p = mt->state;
	q = mt->output;
	for (i = 0; i < N; i++) {
		uint32 y;
		y = *p++;
		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9D2C5680;
		y ^= (y << 15) & 0xEFC60000;
		y ^= (y >> 18);
		*q++ = y;
	}
}

void myMT_SetSeed(myMT_t *mt, uint32 seed)
{
	uint32 *p;
	int i;

	p = mt->state;
	p[0] = seed;

	for (i = 1; i < N; i++) {
		p[i] = (1812433253 * (p[i-1] ^ (p[i-1] >> 30)) + i); 
		/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
		/* In the previous versions, MSBs of the seed affect   */
		/* only MSBs of the array state[]. */
		/* 2002/01/09 modified by Makoto Matsumoto */
	}
	myMT_Next(mt);
}

void myMT_SetSeeds(myMT_t *mt, uint32 *seed, int seedNum)
{
	uint32 *p = mt->state;
	int i;
	int num = (seedNum < N) ? seedNum : N;

	if (num <= 0) {
		myMT_SetSeed(mt, 1234);
		return;
	}

	memcpy(p, seed, num * sizeof(uint32));

	for (i = num; i < N; i++) {
		p[i] = (1812433253 * (p[i-1] ^ (p[i-1] >> 30)) + i); 
		/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
		/* In the previous versions, MSBs of the seed affect   */
		/* only MSBs of the array state[]. */
		/* 2002/01/09 modified by Makoto Matsumoto */
	}
	myMT_Next(mt);
}

/* initializes glb_myMT.state[N] with a seed */
myMT_t *myMT_Init(void)
{
	char *mallocPtr;
	myMT_t *mt;
	const int alignMask = 15;	/* = 2^n - 1 */
	int unit;

	mallocPtr = (char *)malloc(sizeof(myMT_t) + alignMask);
	if (mallocPtr == NULL) return NULL;

	mt = (myMT_t *)(((int)mallocPtr + alignMask) & ~alignMask);
	mt->mallocPtr = mallocPtr;

	unit = haveX86_UNIT();
	if (unit & MU_tSSE2) {
		myMT_Next = myMT_Next_SSE2;
//		myMT_Next = myMT_Next_other;
//		myMT_Next = myMT_Next_MMX;
	} else if (unit & MU_tMMX) {
		myMT_Next = myMT_Next_MMX;
	} else {
		myMT_Next = myMT_Next_C;
	}
	return mt;
}

void myMT_Term(myMT_t *mt)
{
	free(mt->mallocPtr);
}
